from typing import List

from shared.model import Gitmoji
from shared.spec import mojis

# global pattern to validate commit messages
PATTERN = (
    r"(?s)"
    r"(?P<type_group>{type_group})"
    r"(?P<scope>(\(\S+\))?!?:)"
    r"(?P<subject>( [^\n\r]+))"
    r"(?P<body>((\n\n.*)|(\s*))?$)"
)


def get_gitmojis() -> List[Gitmoji]:
    """Return the list of Gitmoji objects."""
    return [Gitmoji(**moji) for moji in mojis]


def get_type_group_pattern() -> str:
    """Return the type group pattern."""
    return "|".join([f"({moji.icon} {{1,2}})?{moji.type}" for moji in get_gitmojis()])


def get_pattern() -> str:
    """Return the complete validation pattern."""
    type_group = get_type_group_pattern()
    return PATTERN.format(type_group=type_group)



__all__ = (
    "GJ_FIX",
    "GJ_FEAT",
    "GJ_DOCS",
    "GJ_STYLE",
    "GJ_REFACTOR",
    "GJ_PERF",
    "GJ_TEST",
    "GJ_BUILD",
    "GJ_CI",
    "GJ_REVERT",
    "GJ_DUMP",
    "GJ_HOTFIX",
    "GJ_DEPLOY",
    "GJ_UI",
    "GJ_INIT",
    "GJ_SECURITY",
    "GJ_SECRET",
    "GJ_BUMP",
    "GJ_FIX_LINT",
    "GJ_WIP",
    "GJ_DEP_DROP",
    "GJ_DEP_BUMP",
    "GJ_PIN",
    "GJ_ANALYTICS",
    "GJ_DEP_ADD",
    "GJ_DEP_RM",
    "GJ_CONFIG",
    "GJ_SCRIPT",
    "GJ_LANG",
    "GJ_TYPO",
    "GJ_POOP",
    "GJ_MERGE",
    "GJ_PACKAGE",
    "GJ_EXTERNAL",
    "GJ_RESOURCE",
    "GJ_LICENSE",
    "GJ_BOOM",
    "GJ_ASSET",
    "GJ_ACCESSIBILITY",
    "GJ_SOURCE_DOCS",
    "GJ_BEER",
    "GJ_TEXT",
    "GJ_DB",
    "GJ_LOGS_ADD",
    "GJ_LOGS_RM",
    "GJ_PEOPLE",
    "GJ_UX",
    "GJ_ARCH",
    "GJ_DESIGN",
    "GJ_MOCK",
    "GJ_EGG",
    "GJ_IGNORE",
    "GJ_SNAP",
    "GJ_EXPERIMENT",
    "GJ_SEO",
    "GJ_TYPES",
    "GJ_SEED",
    "GJ_FLAG",
    "GJ_CATCH",
    "GJ_ANIMATION",
    "GJ_DEPRECATION",
    "GJ_AUTH",
    "GJ_FIX_SIMPLE",
    "GJ_EXPLORATION",
    "GJ_DEAD",
    "GJ_TEST_FAIL",
    "GJ_LOGIC",
    "GJ_HEALTH",
    "GJ_INFRA",
    "GJ_DEVXP",
    "GJ_MONEY",
    "GJ_THREADING",
    "GJ_VALIDATION",
    "GJ_CHORE",
    "GJ_SQUASH",
    "GJ_FIXUP",
)

GJ_FIX = "🐛"
GJ_FEAT = "✨"
GJ_DOCS = "📝"
GJ_STYLE = "🎨"
GJ_REFACTOR = "♻️ "
GJ_PERF = "⚡️"
GJ_TEST = "✅"
GJ_BUILD = "👷"
GJ_CI = "💚"
GJ_REVERT = "⏪️"
GJ_DUMP = "🔥"
GJ_HOTFIX = "🚑️"
GJ_DEPLOY = "🚀"
GJ_UI = "💄"
GJ_INIT = "🎉"
GJ_SECURITY = "🔒️"
GJ_SECRET = "🔐"
GJ_BUMP = "🔖"
GJ_FIX_LINT = "🚨"
GJ_WIP = "🚧"
GJ_DEP_DROP = "⬇️ "
GJ_DEP_BUMP = "⬆️ "
GJ_PIN = "📌"
GJ_ANALYTICS = "📈"
GJ_DEP_ADD = "➕"
GJ_DEP_RM = "➖"
GJ_CONFIG = "🔧"
GJ_SCRIPT = "🔨"
GJ_LANG = "🌐"
GJ_TYPO = "✏️ "
GJ_POOP = "💩"
GJ_MERGE = "🔀"
GJ_PACKAGE = "📦️"
GJ_EXTERNAL = "👽️"
GJ_RESOURCE = "🚚"
GJ_LICENSE = "📄"
GJ_BOOM = "💥"
GJ_ASSET = "🍱"
GJ_ACCESSIBILITY = "♿️"
GJ_SOURCE_DOCS = "💡"
GJ_BEER = "🍻"
GJ_TEXT = "💬"
GJ_DB = "🗃️"
GJ_LOGS_ADD = "🔊"
GJ_LOGS_RM = "🔇"
GJ_PEOPLE = "👥"
GJ_UX = "🚸"
GJ_ARCH = "🏗️"
GJ_DESIGN = "📱"
GJ_MOCK = "🤡"
GJ_EGG = "🥚"
GJ_IGNORE = "🙈"
GJ_SNAP = "📸"
GJ_EXPERIMENT = "⚗️ "
GJ_SEO = "🔍️"
GJ_TYPES = "🏷️"
GJ_SEED = "🌱"
GJ_FLAG = "🚩"
GJ_CATCH = "🥅"
GJ_ANIMATION = "💫"
GJ_DEPRECATION = "🗑️"
GJ_AUTH = "🛂"
GJ_FIX_SIMPLE = "🩹"
GJ_EXPLORATION = "🧐"
GJ_DEAD = "⚰️ "
GJ_TEST_FAIL = "🧪"
GJ_LOGIC = "👔"
GJ_HEALTH = "🩺"
GJ_INFRA = "🧱"
GJ_DEVXP = "🧑‍💻"
GJ_MONEY = "💸"
GJ_THREADING = "🧵"
GJ_VALIDATION = "🦺"
GJ_CHORE = "🧹"
GJ_SQUASH = "👇"
GJ_FIXUP = "🫥"



import argparse
import re
import sys
from pathlib import Path
from typing import Dict, List, Optional

from shared.model import Gitmoji
from shared.settings import get_settings
from shared.utils import get_gitmojis, get_pattern


def _get_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-f",
        "--commit-msg-file",
        help="path to the commit message file",
    )
    group.add_argument("-m", "--message", help="the commit message")
    parser.add_argument("--config", help="path to the configuration file")
    parser.add_argument(
        "--allowed-prefixes",
        nargs="*",
        help="allowed gitmoji prefixes",
    )
    parser.add_argument(
        "--convert-prefixes",
        nargs="*",
        help="prefixes to convert to gitmoji format",
    )
    return parser.parse_args()


def _grouped_gitmojis() -> Dict[str, Gitmoji]:
    """Return the gitmojis grouped by type."""
    return {moji.type: moji for moji in get_gitmojis()}


def gitmojify(
    message: str,
    allowed_prefixes: Optional[List[str]] = None,
    convert_prefixes: Optional[List[str]] = None,
) -> str:
    """
    Gitmojify the commit message.

    If a gitmoji is already present in the message, the message is returned
    as is. Otherwise, the gitmoji is looked up by type and is prepended to
    the message.

    Args:
        message: The complete commit message.
        allowed_prefixes: Prefixes that should not raise an error, even though
            they're not following conventional standard.
        convert_prefixes: Prefixes that should be converted to gitmoji format.

    Returns:
        The gitmojified message.
    """
    convert_prefixes = convert_prefixes or []
    if any(map(message.startswith, convert_prefixes)):
        first_word, *rest = message.split(" ", maxsplit=1)
        if first_word.endswith(":"):
            first_word = first_word[:-1]
        if first_word in convert_prefixes:
            message = f"{first_word.lower()}: {''.join(rest)}"
    if any(map(message.startswith, allowed_prefixes or [])):
        return message

    pat = get_pattern()
    match = re.match(pat, message)
    if match is None:
        msg = "invalid commit message"
        raise ValueError(msg)
    gtype = match.group("type_group")
    if " " in gtype:  # maybe do a better check?
        return message
    icon = _grouped_gitmojis()[gtype].icon
    return f"{icon} {message}"


def _write(filepath: Optional[Path], message: str, encoding: str) -> None:
    """Write the message to the file."""
    if filepath is None:
        sys.stdout.write(message)
        return
    with filepath.open("w", encoding=encoding) as f:
        f.write(message)


def _filter_comments(message: str) -> str:
    """Filter out comments from the message.

    Copied from https://github.com/commitizen-tools/commitizen/blob/a54bf542012257a338edb54ca16d11ba6beef0ba/commitizen/commands/check.py#L113-L141
    """
    lines = []
    for line in message.split("\n"):
        if "# ------------------------ >8 ------------------------" in line:
            break
        if not line.startswith("#"):
            lines.append(line)
    return "\n".join(lines)


def run() -> None:
    """The pre-commit hook that modifies the commit message."""
    args = _get_args()
    settings = get_settings(args.config)
    if args.commit_msg_file:
        filepath = Path(args.commit_msg_file)
        msg = filepath.read_text(encoding=settings.encoding)
    else:
        filepath = None
        msg = args.message
    _write(
        filepath,
        gitmojify(
            _filter_comments(msg),
            args.allowed_prefixes or settings.allowed_prefixes,
            args.convert_prefixes or settings.convert_prefixes,
        ),
        settings.encoding,
    )



"""Settings."""

from typing import List, Optional

import attrs
from commitizen import config

DEFAULT_CONVERT_PREFIXES = ["Merge", "Revert", "Squash"]


@attrs.define(kw_only=True)
class MojiSettings:
    """Settings."""

    allowed_prefixes: List[str]
    convert_prefixes: List[str] = attrs.field(factory=lambda: DEFAULT_CONVERT_PREFIXES)
    conventional_types_only: bool = False
    conventional_messages: bool = False
    encoding: str


def get_settings(filepath: Optional[str] = None) -> MojiSettings:
    """Get settings."""
    cfg = config.read_cfg(filepath)
    fields_used = {field.name for field in attrs.fields(MojiSettings)}
    return MojiSettings(**{k: v for k, v in cfg.settings.items() if k in fields_used})  # type: ignore[arg-type]
